# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import Any, Dict, List, Tuple

import tensorflow as tf

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.common.compression import BaseCompressionAlgorithmController
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import ConvertDtypeLayerAttributes
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OUTPUT_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.graph.utils import get_first_nodes_of_type
from nncf.common.hardware.config import get_hw_config_type
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.insertion_point_graph import InsertionPointGraph
from nncf.common.logging import nncf_logger
from nncf.common.quantization.config_assignment import assign_qconfig_lists_to_modules
from nncf.common.quantization.quantizer_propagation.solver import QuantizerPropagationSolver
from nncf.common.quantization.quantizer_propagation.structs import IgnoreReason
from nncf.common.quantization.quantizer_setup import ActivationQuantizationInsertionPoint
from nncf.common.quantization.quantizer_setup import QuantizationPointId
from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.common.quantization.structs import QuantizableWeightedLayerNode
from nncf.common.quantization.structs import QuantizationConstraints
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import QuantizerGroup
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.stateful_classes_registry import TF_STATEFUL_CLASSES
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import copy_model
from nncf.config.extractors import extract_range_init_params
from nncf.config.schemata.defaults import QUANTIZATION_OVERFLOW_FIX
from nncf.config.schemata.defaults import QUANTIZE_INPUTS
from nncf.config.schemata.defaults import QUANTIZE_OUTPUTS
from nncf.config.schemata.defaults import TARGET_DEVICE
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.converter import TFModelConverter
from nncf.tensorflow.graph.converter import TFModelConverterFactory
from nncf.tensorflow.graph.metatypes.common import CAST_METATYPES
from nncf.tensorflow.graph.metatypes.common import ELEMENTWISE_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import GENERAL_CONV_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.common import LINEAR_LAYER_METATYPES
from nncf.tensorflow.graph.metatypes.keras_layers import TFConcatenateLayerMetatype
from nncf.tensorflow.graph.metatypes.keras_layers import TFLambdaLayerMetatype
from nncf.tensorflow.graph.metatypes.keras_layers import TFLayerWithWeightsMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFConcatOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFIdentityOpMetatype
from nncf.tensorflow.graph.metatypes.tf_ops import TFOpWithWeightsMetatype
from nncf.tensorflow.graph.transformations.commands import TFAfterLayer
from nncf.tensorflow.graph.transformations.commands import TFBeforeLayer
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.commands import TFMultiLayerPoint
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx
from nncf.tensorflow.hardware.config import TFHWConfig
from nncf.tensorflow.hardware.fused_patterns import TF_HW_FUSED_PATTERNS
from nncf.tensorflow.layers.custom_objects import NNCF_QUANTIZATION_OPERATIONS
from nncf.tensorflow.loss import TFZeroCompressionLoss
from nncf.tensorflow.quantization.collectors import TFQuantizationStatisticsCollector
from nncf.tensorflow.quantization.default_quantization import DEFAULT_TF_QUANT_TRAIT_TO_OP_DICT
from nncf.tensorflow.quantization.init_range import RangeInitializer
from nncf.tensorflow.quantization.init_range import TFRangeInitParams
from nncf.tensorflow.quantization.layers import FakeQuantize
from nncf.tensorflow.quantization.quantizers import Quantizer
from nncf.tensorflow.quantization.quantizers import TFQuantizerSpec
from nncf.tensorflow.quantization.utils import apply_overflow_fix

QUANTIZATION_LAYER_METATYPES = GENERAL_CONV_LAYER_METATYPES + LINEAR_LAYER_METATYPES

UNSUPPORTED_LAYER_METATYPES = [TFLambdaLayerMetatype]


class TFQuantizationPointStateNames:
    QUANTIZER_SPEC = "quantizer_spec"
    TARGET_POINT = "target_point"
    TARGET_POINT_CLASS_NAME = "target_point_class_name"
    OP_NAME = "op_name"


class TFQuantizationPoint:
    """
    Characterizes where and how to insert a single quantization node to the model's graph. Stores TF-specific data.
    """

    _state_names = TFQuantizationPointStateNames

    def __init__(self, op_name: str, quantizer_spec: TFQuantizerSpec, target_point: TargetPoint):
        self.target_point = target_point
        self.op_name = op_name
        self.quantizer_spec = quantizer_spec

    def is_weight_quantization(self) -> bool:
        """
        Determines whether quantization point is for weights
        :return: true, if quantization for weights, false - for activations
        """
        return isinstance(self.target_point, TFLayerWeight)

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.TARGET_POINT: self.target_point.get_state(),
            self._state_names.TARGET_POINT_CLASS_NAME: self.target_point.__class__.__name__,
            self._state_names.QUANTIZER_SPEC: self.quantizer_spec.get_state(),
            self._state_names.OP_NAME: self.op_name,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "TFQuantizationPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        target_point_cls = TF_STATEFUL_CLASSES.get_registered_class(state[cls._state_names.TARGET_POINT_CLASS_NAME])
        kwargs = {
            cls._state_names.TARGET_POINT: target_point_cls.from_state(state[cls._state_names.TARGET_POINT]),
            cls._state_names.QUANTIZER_SPEC: TFQuantizerSpec.from_state(state[cls._state_names.QUANTIZER_SPEC]),
            cls._state_names.OP_NAME: state[cls._state_names.OP_NAME],
        }
        return cls(**kwargs)


class TFQuantizationSetupStateNames:
    QUANTIZATION_POINTS = "quantization_points"
    UNIFIED_SCALE_GROUPS = "unified_scale_groups"


class TFQuantizationSetup:
    """
    Characterizes where and how to insert all quantization nodes to the model's graph
    """

    _state_names = TFQuantizationSetupStateNames

    def __init__(self):
        super().__init__()
        self._quantization_points: List[TFQuantizationPoint] = []
        self._unified_scale_groups: List[List[QuantizationPointId]] = []

    def add_quantization_point(self, point: TFQuantizationPoint):
        """
        Adds quantization point to the setup

        :param point: quantization point
        """
        self._quantization_points.append(point)

    def __iter__(self):
        return iter(self._quantization_points)

    def register_unified_scale_group(self, point_ids: List[List[QuantizationPointId]]):
        """
        Adds unified scale group to the setup

        :param point_ids: quantization point indexes
        """
        self._unified_scale_groups.append(point_ids)

    def get_quantization_points(self) -> List[TFQuantizationPoint]:
        """
        Returns quantization points

        :return: quantization points
        """
        return self._quantization_points

    def get_unified_scale_groups(self) -> List[List[QuantizationPointId]]:
        """
        Returns unified scale groups

        :return: unified scale groups
        """
        return self._unified_scale_groups

    def get_state(self) -> Dict:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """

        quantization_points_state = [qp.get_state() for qp in self._quantization_points]
        return {
            self._state_names.QUANTIZATION_POINTS: quantization_points_state,
            self._state_names.UNIFIED_SCALE_GROUPS: self._unified_scale_groups,
        }

    @classmethod
    def from_state(cls, state: Dict) -> "TFQuantizationSetup":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        setup = TFQuantizationSetup()
        for quantization_point_state in state[cls._state_names.QUANTIZATION_POINTS]:
            quantization_point = TFQuantizationPoint.from_state(quantization_point_state)
            setup.add_quantization_point(quantization_point)

        if cls._state_names.UNIFIED_SCALE_GROUPS in state:
            for quantization_group in state[cls._state_names.UNIFIED_SCALE_GROUPS]:
                setup.register_unified_scale_group(quantization_group)
        return setup


class QBuilderStateNames:
    QUANTIZER_SETUP = "quantizer_setup"


@TF_COMPRESSION_ALGORITHMS.register("quantization")
class QuantizationBuilder(TFCompressionAlgorithmBuilder):
    _state_names = QBuilderStateNames

    DEFAULT_QCONFIG = QuantizerConfig(
        num_bits=8, mode=QuantizationMode.SYMMETRIC, signedness_to_force=None, per_channel=False
    )

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)

        self.quantize_inputs = self._algo_config.get("quantize_inputs", QUANTIZE_INPUTS)
        self.quantize_outputs = self._algo_config.get("quantize_outputs", QUANTIZE_OUTPUTS)
        self._overflow_fix = self._algo_config.get("overflow_fix", QUANTIZATION_OVERFLOW_FIX)
        self._target_device = config.get("target_device", TARGET_DEVICE)
        algo_config = self._get_algo_specific_config_section()
        if self._target_device == "NPU" and "preset" in algo_config:
            raise nncf.ValidationError("The NPU target device does not support presets.")
        if self._target_device == "CPU_SPR":
            raise nncf.ValidationError("The CPU_SPR target device does not supported.")

        self.global_quantizer_constraints = {}
        self.ignored_scopes_per_group = {}
        self.target_scopes_per_group = {}
        self._op_names = []

        for quantizer_group in QuantizerGroup:
            self._parse_group_params(self._algo_config, quantizer_group)

        if self.should_init:
            self._parse_init_params()

        self._range_initializer = None
        self._bn_adaptation = None
        self._quantizer_setup = None

        self.hw_config = None
        hw_config_type = get_hw_config_type(self._target_device)
        if hw_config_type is not None:
            hw_config_path = TFHWConfig.get_path_to_hw_config(hw_config_type)
            self.hw_config = TFHWConfig.from_json(hw_config_path)

    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Initializes object from the state.

        :param state_without_name: Output of `get_state()` method.
        """
        quantizer_setup_state = state_without_name[self._state_names.QUANTIZER_SETUP]
        self._quantizer_setup = TFQuantizationSetup.from_state(quantizer_setup_state)

    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        quantizer_setup_state = self._quantizer_setup.get_state()
        return {self._state_names.QUANTIZER_SETUP: quantizer_setup_state}

    def _parse_init_params(self):
        self._range_init_params = self._parse_range_init_params()
        self._bn_adapt_params = self._parse_bn_adapt_params()

    def _parse_range_init_params(self) -> TFRangeInitParams:
        range_init_params = extract_range_init_params(self.config)
        return TFRangeInitParams(**range_init_params) if range_init_params is not None else None

    def _parse_group_params(self, quant_config: Dict, quantizer_group: QuantizerGroup) -> None:
        group_name = quantizer_group.value
        params_dict = {}
        params_dict_from_config = quant_config.get(group_name, {})
        preset = quant_config.get("preset")
        if self._target_device in ["ANY", "CPU", "GPU"] or self._target_device == "TRIAL" and preset is not None:
            preset = QuantizationPreset(quant_config.get("preset", "performance"))
            params_dict = preset.get_params_configured_by_preset(quantizer_group)
            overriden_params = params_dict.keys() & params_dict_from_config.keys()
            if overriden_params:
                nncf_logger.info(f"Preset quantizer parameters {overriden_params} explicitly overridden by config.")
        params_dict.update(params_dict_from_config)
        self.global_quantizer_constraints[quantizer_group] = QuantizationConstraints.from_config_dict(params_dict)
        self.ignored_scopes_per_group[quantizer_group] = params_dict_from_config.get("ignored_scopes", [])
        if self.ignored_scopes is not None:
            self.ignored_scopes_per_group[quantizer_group] += self.ignored_scopes
        target_scopes = params_dict_from_config.get("target_scopes")
        if target_scopes is None and self.target_scopes is not None:
            self.target_scopes_per_group[quantizer_group] = self.target_scopes
        else:
            self.target_scopes_per_group[quantizer_group] = target_scopes

    def _get_default_qconfig(self, constraints: QuantizationConstraints = None) -> QuantizerConfig:
        qconfig = deepcopy(self.DEFAULT_QCONFIG)
        if constraints is not None:
            qconfig = constraints.apply_constraints_to(qconfig)
        return qconfig

    def _get_half_range(
        self, qconfig: QuantizerConfig, target_node: NNCFNode, first_conv_nodes: List[NNCFNode]
    ) -> bool:
        if self._target_device in ["CPU", "ANY"] and qconfig.num_bits == 8:
            if self._overflow_fix == "enable":
                return True
            if self._overflow_fix == "first_layer_only" and target_node in first_conv_nodes:
                return True
        return False

    def _create_quantizer(self, name: str, qspec: TFQuantizerSpec) -> Quantizer:
        quantizer_cls = NNCF_QUANTIZATION_OPERATIONS.get(qspec.mode)
        return quantizer_cls(name, qspec)

    def _build_insertion_commands_for_quantizer_setup(
        self, quantizer_setup: TFQuantizationSetup
    ) -> List[TFInsertionCommand]:
        insertion_commands = []
        quantization_points = quantizer_setup.get_quantization_points()
        non_unified_scales_quantization_point_ids = set(range(len(quantization_points)))

        for unified_scales_group in quantizer_setup.get_unified_scale_groups():
            us_qp_id = unified_scales_group[0]
            qp = quantization_points[us_qp_id]
            quantizer_spec = qp.quantizer_spec
            op_name = qp.op_name + "/unified_scale_group"
            quantizer = FakeQuantize(quantizer_spec, name=op_name)
            self._op_names.append(quantizer.op_name)
            target_points = []
            for us_qp_id in unified_scales_group:
                non_unified_scales_quantization_point_ids.discard(us_qp_id)
                qp = quantization_points[us_qp_id]
                assert quantizer_spec.get_state() == qp.quantizer_spec.get_state()
                target_points.append(qp.target_point)

            command = TFInsertionCommand(
                target_point=TFMultiLayerPoint(target_points),
                callable_object=quantizer,
                priority=TransformationPriority.QUANTIZATION_PRIORITY,
            )

            insertion_commands.append(command)

        for qp_id in non_unified_scales_quantization_point_ids:
            quantization_point = quantization_points[qp_id]
            op_name = quantization_point.op_name
            quantizer_spec = quantization_point.quantizer_spec
            target_point = quantization_point.target_point
            if quantization_point.is_weight_quantization():
                quantizer = self._create_quantizer(op_name, quantizer_spec)
                self._op_names.append(op_name)
            else:
                quantizer = FakeQuantize(quantizer_spec, name=op_name)
                self._op_names.append(quantizer.op_name)
            command = TFInsertionCommand(
                target_point=target_point,
                callable_object=quantizer,
                priority=TransformationPriority.QUANTIZATION_PRIORITY,
            )
            insertion_commands.append(command)
        return insertion_commands

    def get_transformation_layout(self, model: tf.keras.Model) -> TFTransformationLayout:
        transformations = TFTransformationLayout()
        if self._quantizer_setup is None:
            self._quantizer_setup = self._get_quantizer_setup(model)
        insertion_commands = self._build_insertion_commands_for_quantizer_setup(self._quantizer_setup)
        for command in insertion_commands:
            transformations.register(command)
        return transformations

    def _get_custom_layer_node_names(self, nncf_graph: NNCFGraph, converter: TFModelConverter) -> List[NNCFNodeName]:
        retval = []
        for node in nncf_graph.get_all_nodes():
            metatype = node.metatype
            if metatype in OUTPUT_NOOP_METATYPES:
                continue
            is_custom, _ = converter.get_layer_info_for_node(node.node_name)
            if is_custom:
                retval.append(node.node_name)
        return retval

    def _build_controller(self, model: tf.keras.Model) -> "QuantizationController":
        return QuantizationController(model, self.config, self._op_names)

    def initialize(self, model: tf.keras.Model) -> None:
        if self._range_init_params is not None:
            self._run_range_initialization(model)
        if self._bn_adapt_params is not None:
            self._run_batchnorm_adaptation(model)

    def _run_range_initialization(self, model: tf.keras.Model) -> None:
        if self._range_initializer is None:
            self._range_initializer = RangeInitializer(self._range_init_params)
        self._range_initializer.run(model)

    def _run_batchnorm_adaptation(self, model: tf.keras.Model) -> None:
        if self._bn_adaptation is None:
            self._bn_adaptation = BatchnormAdaptationAlgorithm(
                self._bn_adapt_params["data_loader"],
                self._bn_adapt_params["num_bn_adaptation_samples"],
                self._bn_adapt_params["device"],
            )
        self._bn_adaptation.run(model)

    def _get_quantizer_setup(self, model: tf.keras.Model) -> TFQuantizationSetup:
        converter = TFModelConverterFactory.create(model)
        nncf_graph = converter.convert()
        nncf_graph = QuantizationBuilder._preprocess_cast_nodes(nncf_graph, CAST_METATYPES)

        check_scopes_in_graph(nncf_graph, self.ignored_scopes, self.target_scopes, self.validate_scopes)

        self._raise_not_supported_warning(nncf_graph)

        quantizable_weighted_layer_nodes = self._get_quantizable_weighted_layer_nodes(nncf_graph)
        custom_layer_nodes = self._get_custom_layer_node_names(nncf_graph, converter)

        quantizer_setup = self._get_quantizer_propagation_solution(
            nncf_graph, quantizable_weighted_layer_nodes, custom_layer_nodes, model
        )
        setup = TFQuantizationSetup()

        quantized_layer_names_vs_qconfigs: Dict[str, QuantizerConfig] = {}
        qp_id_to_index: Dict[QuantizationPointId, int] = {}
        tf_setup_qp_index = 0
        applied_overflow_fix = False
        first_conv_nodes = get_first_nodes_of_type(nncf_graph, ["Conv2D", "Conv3D"])
        for qp_id, qp in quantizer_setup.quantization_points.items():
            if qp.is_weight_quantization_point():
                target_node = nncf_graph.get_node_by_name(qp.insertion_point.target_node_name)
                is_custom, layer_info = converter.get_layer_info_for_node(target_node.node_name)
                if is_custom:
                    raise nncf.InternalError("Quantizing custom layer weights is currently unsupported!")
                layer_name = layer_info.layer_name
                qconfig = qp.qconfig
                if layer_name in quantized_layer_names_vs_qconfigs:
                    assigned_qconfig = quantized_layer_names_vs_qconfigs[layer_name]
                    if qconfig != assigned_qconfig:
                        raise nncf.InternalError(
                            f"Inconsistent quantizer configurations selected by solver for one and the "
                            f"same quantizable layer! Tried to assign {qconfig} to {layer_name} as "
                            f"specified by QP {qp_id}, but the layer already has quantizer "
                            f"config {assigned_qconfig} assigned to it!"
                        )
                    continue  # The layer has already been quantized
                quantized_layer_names_vs_qconfigs[layer_name] = qconfig
                metatype = target_node.metatype
                assert issubclass(metatype, TFLayerWithWeightsMetatype)
                for weight_def in metatype.weight_definitions:
                    op_name = self._get_quantizer_operation_name(target_node.node_name, weight_def.weight_attr_name)
                    self._op_names.append(op_name)

                    half_range = self._get_half_range(qconfig, target_node, first_conv_nodes)
                    applied_overflow_fix = applied_overflow_fix or half_range
                    quantizer_spec = TFQuantizerSpec.from_config(
                        qconfig, narrow_range=not half_range, half_range=half_range
                    )
                    target_point = TFLayerWeight(layer_info.layer_name, weight_def.weight_attr_name)
                    qpoint = TFQuantizationPoint(op_name, quantizer_spec, target_point)
            else:
                assert qp.is_activation_quantization_point()
                ip = qp.insertion_point
                assert isinstance(ip, ActivationQuantizationInsertionPoint)
                target_node_name = ip.target_node_name
                input_port_id = ip.input_port_id
                fake_quantize_name = self._get_fake_quantize_name(target_node_name, input_port_id)
                quantizer_spec = TFQuantizerSpec.from_config(qp.qconfig, narrow_range=False, half_range=False)
                fake_quantize_layer = FakeQuantize(quantizer_spec, name=fake_quantize_name)
                self._op_names.append(fake_quantize_layer.op_name)

                is_custom, layer_info = converter.get_layer_info_for_node(target_node_name)
                if is_custom:
                    raise nncf.InternalError("Quantizing custom layer activations is currently unsupported!")
                if input_port_id is not None:
                    target_point = TFBeforeLayer(
                        layer_info.layer_name, instance_idx=layer_info.instance_idx, input_port_id=input_port_id
                    )
                else:
                    target_point = TFAfterLayer(
                        layer_info.layer_name, instance_idx=layer_info.instance_idx, output_port_id=0
                    )
                qpoint = TFQuantizationPoint(fake_quantize_name, quantizer_spec, target_point)

            setup.add_quantization_point(qpoint)
            qp_id_to_index[qp_id] = tf_setup_qp_index
            tf_setup_qp_index += 1

        setup = self._generate_unified_scale_groups(model, quantizer_setup, qp_id_to_index, setup)

        self._log_if_overflow_fix_was_applied(applied_overflow_fix)

        return setup

    def _raise_not_supported_warning(self, graph: NNCFGraph) -> None:
        for node in graph.get_all_nodes():
            if node.metatype in UNSUPPORTED_LAYER_METATYPES:
                nncf_logger.warning(
                    f"Layer "
                    f"{get_original_name_and_instance_idx(node.node_name)[0]} "
                    f"is not supported by the quantization algorithm"
                )

    def _log_if_overflow_fix_was_applied(self, applied_overflow_fix: bool):
        if applied_overflow_fix:
            if self._overflow_fix == "enable":
                quantizers_with_overflow_fix_str = "all weight quantizers"
            elif self._overflow_fix == "first_layer_only":
                quantizers_with_overflow_fix_str = "first convolution weight quantizers"
            elif self._overflow_fix != "disable":
                raise nncf.InternalError(f"Unknown overflow fix type: {self._overflow_fix}")
            nncf_logger.info(f"Overflow issue fix was applied to {quantizers_with_overflow_fix_str}.")

    def _generate_unified_scale_groups(
        self,
        model: tf.keras.Model,
        quantizer_setup: SingleConfigQuantizerSetup,
        qp_id_to_index: Dict[QuantizationPointId, int],
        setup: TFQuantizationSetup,
    ) -> TFQuantizationSetup:
        # To properly set the instance indices for FQ need to save layers order like in the model config
        layer_names = [layer.name for layer in model.layers]
        for unified_group in quantizer_setup.unified_scale_groups.values():
            sorted_unified_group = []
            for qp_id in unified_group:
                qp = quantizer_setup.quantization_points[qp_id]
                qp_layer_name = qp.insertion_point.target_node_name
                original_name, _ = get_original_name_and_instance_idx(qp_layer_name)
                layer_idx = layer_names.index(original_name)
                tf_setup_index = qp_id_to_index[qp_id]
                sorted_unified_group.append((tf_setup_index, layer_idx))

            sorted_unified_group = sorted(sorted_unified_group, key=lambda x: x[1])
            setup.register_unified_scale_group([setup_index for setup_index, _ in sorted_unified_group])
        return setup

    def _get_quantizable_weighted_layer_nodes(self, nncf_graph: NNCFGraph) -> List[QuantizableWeightedLayerNode]:
        nodes_with_weights = []
        for node in nncf_graph.get_all_nodes():
            metatype = node.metatype
            if metatype in OUTPUT_NOOP_METATYPES:
                continue

            if metatype not in QUANTIZATION_LAYER_METATYPES:
                continue

            assert issubclass(metatype, TFLayerWithWeightsMetatype) or issubclass(metatype, TFOpWithWeightsMetatype)
            nodes_with_weights.append(node)
        scope_overrides_dict = self._get_algo_specific_config_section().get("scope_overrides", {})
        weighted_node_and_qconf_lists = assign_qconfig_lists_to_modules(
            nodes_with_weights,
            self.DEFAULT_QCONFIG,
            self.global_quantizer_constraints[QuantizerGroup.WEIGHTS],
            scope_overrides_dict,
            hw_config=self.hw_config,
        )
        return [
            QuantizableWeightedLayerNode(node, qconf_list) for node, qconf_list in weighted_node_and_qconf_lists.items()
        ]

    def _get_quantizer_propagation_solution(
        self,
        nncf_graph: NNCFGraph,
        quantizable_weighted_layer_nodes: List[QuantizableWeightedLayerNode],
        custom_layer_node_names: List[NNCFNodeName],
        model: tf.keras.Model,
    ) -> SingleConfigQuantizerSetup:
        ip_graph = InsertionPointGraph(nncf_graph, [qn.node.node_name for qn in quantizable_weighted_layer_nodes])

        pattern = TF_HW_FUSED_PATTERNS.get_full_pattern_graph()
        ip_graph = ip_graph.get_ip_graph_with_merged_hw_optimized_operations(pattern)

        input_preprocessing_nodes = self._get_input_preprocessing_nodes(nncf_graph, model)
        input_preprocessing_node_names = [n.node_name for n in input_preprocessing_nodes]
        if custom_layer_node_names:
            custom_layer_node_names_str = ", ".join([str(n) for n in custom_layer_node_names])
            nncf_logger.warning(
                f"Following custom layers will be ignored during quantization (custom layer quantization not supported "
                f"by NNCF yet):\n[{custom_layer_node_names_str}]"
            )
        ignored_scopes_for_solver = {
            **{name: IgnoreReason.USER_REQUESTED for name in self.ignored_scopes_per_group[QuantizerGroup.ACTIVATIONS]},
            **{name: IgnoreReason.AUTOGENERATED for name in input_preprocessing_node_names},
            **{name: IgnoreReason.AUTOGENERATED for name in custom_layer_node_names},
        }
        scales_unification_map = {
            TFConcatenateLayerMetatype: GENERAL_CONV_LAYER_METATYPES + LINEAR_LAYER_METATYPES,
            TFConcatOpMetatype: GENERAL_CONV_LAYER_METATYPES + LINEAR_LAYER_METATYPES,
        }
        solver = QuantizerPropagationSolver(
            activation_ignored_scopes=ignored_scopes_for_solver,
            weight_ignored_scopes=self.ignored_scopes_per_group[QuantizerGroup.WEIGHTS],
            activation_target_scopes=self.target_scopes_per_group[QuantizerGroup.ACTIVATIONS],
            weight_target_scopes=self.target_scopes_per_group[QuantizerGroup.WEIGHTS],
            hw_config=self.hw_config,
            default_trait_to_metatype_map=DEFAULT_TF_QUANT_TRAIT_TO_OP_DICT,
            default_qconfig_list=[
                self._get_default_qconfig(self.global_quantizer_constraints[QuantizerGroup.ACTIVATIONS])
            ],
            quantizable_layer_nodes=quantizable_weighted_layer_nodes,
            global_constraints=self.global_quantizer_constraints,
            quantize_outputs=self.quantize_outputs,
            scales_unification_map=scales_unification_map,
        )

        quantization_proposal = solver.run_on_ip_graph(ip_graph)
        multi_config_setup = quantization_proposal.quantizer_setup
        single_config_setup = multi_config_setup.select_first_qconfig_for_each_point()
        finalized_proposal = quantization_proposal.finalize(single_config_setup)
        final_setup = solver.get_final_quantizer_setup(finalized_proposal)
        final_setup = self._handle_quantize_inputs_option(final_setup, nncf_graph)

        return final_setup

    def _handle_quantize_inputs_option(
        self, quantizer_setup: SingleConfigQuantizerSetup, nncf_graph: NNCFGraph
    ) -> SingleConfigQuantizerSetup:
        qp_ids_to_discard = []
        for qp_id, qp in quantizer_setup.quantization_points.items():
            if qp.is_activation_quantization_point():
                insertion_point = qp.insertion_point
                target_node = nncf_graph.get_node_by_name(insertion_point.target_node_name)
                if not self.quantize_inputs and target_node.metatype in INPUT_NOOP_METATYPES:
                    qp_ids_to_discard.append(qp_id)
        for qp_id in qp_ids_to_discard:
            quantizer_setup.discard(qp_id, keep_shared_input_qps=True)
        return quantizer_setup

    def _get_input_preprocessing_nodes(self, nncf_graph: NNCFGraph, model: tf.keras.Model) -> List[NNCFNode]:
        retval = []

        def traverse_fn(node: NNCFNode, preprocessing_nodes: List[NNCFNode]) -> Tuple[bool, List[NNCFNode]]:
            is_finished = True
            successors = nncf_graph.get_next_nodes(node)
            if len(successors) == 1:
                successor = next(iter(successors))
                # It is necessary to determine the number of input nodes from the model
                # in order to correctly count the duplicated edges
                original_name, _ = get_original_name_and_instance_idx(successor.node_name)
                layer = model.get_layer(name=original_name)

                num_previous_nodes = len(layer.input) if isinstance(layer.input, list) else 1
                if layer.__class__.__name__ in ["TFOpLambda", "SlicingOpLambda"]:
                    num_previous_nodes = 0
                    for inbound_node in layer.inbound_nodes:
                        num_previous_nodes += len(inbound_node.keras_inputs)

                if successor.metatype in ELEMENTWISE_LAYER_METATYPES and num_previous_nodes == 1:
                    preprocessing_nodes.append(successor)
                    is_finished = False
            return is_finished, preprocessing_nodes

        for nncf_node in nncf_graph.get_input_nodes():
            preprocessing_nodes_for_this_input = nncf_graph.traverse_graph(nncf_node, traverse_fn)
            retval += preprocessing_nodes_for_this_input

        return retval

    @staticmethod
    def _preprocess_cast_nodes(nncf_graph: NNCFGraph, cast_metatypes: List[OperatorMetatype]) -> NNCFGraph:
        cast_nodes = nncf_graph.get_nodes_by_metatypes(cast_metatypes)
        for node in cast_nodes:
            if not isinstance(node.layer_attributes, ConvertDtypeLayerAttributes):
                continue
            if node.layer_attributes.src_dtype == node.layer_attributes.dst_dtype:
                node.attributes[NNCFNode.METATYPE_ATTR] = TFIdentityOpMetatype
        return nncf_graph

    def _get_fake_quantize_name(self, node_name: NNCFNodeName, input_port_id: int = None) -> str:
        original_node_name, instance_idx = get_original_name_and_instance_idx(node_name)
        fq_name = "{}/fake_quantize".format(original_node_name)
        if instance_idx != 0:
            fq_name += f"_{instance_idx}"
        if input_port_id is not None:
            fq_name += f"_I{input_port_id}"
        return fq_name

    def _get_quantizer_operation_name(self, layer_name, weight_attr_name):
        return f"{layer_name}_{weight_attr_name}_quantizer"


@api()
class QuantizationController(BaseCompressionAlgorithmController):
    """
    Controller for the quantization algorithm in TensorFlow.
    """

    def __init__(self, target_model, config: NNCFConfig, op_names: List[str]):
        super().__init__(target_model)
        self._scheduler = BaseCompressionScheduler()
        self._loss = TFZeroCompressionLoss()
        self._op_names = op_names
        self._config = config

    @property
    def scheduler(self) -> CompressionScheduler:
        return self._scheduler

    @property
    def loss(self) -> CompressionLoss:
        """
        Returns the loss that is always zero since the quantization algorithm is driven by the original loss and does
        not require additional losses.
        """
        return self._loss

    def strip_model(self, model: tf.keras.Model, do_copy: bool = False) -> tf.keras.Model:
        if do_copy:
            model = copy_model(model)
        apply_overflow_fix(model, self._op_names)
        return model

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = TFQuantizationStatisticsCollector(self.model, self._op_names)
        stats = collector.collect()

        nncf_stats = NNCFStatistics()
        nncf_stats.register("quantization", stats)
        return nncf_stats

    def compression_stage(self) -> CompressionStage:
        return CompressionStage.FULLY_COMPRESSED



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import torch
from torch.quantization.fake_quantize import FakeQuantize

import nncf
from nncf.torch.graph.transformations.commands import ExtraCompressionModuleType
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import AsymmetricQuantizer
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import SymmetricQuantizer

SUPPORTED_NUM_BITS_FOR_STRIP_MODEL = [8]


def replace_quantizer_to_torch_native_module(model: NNCFNetwork) -> NNCFNetwork:
    """
    Replace NNCF quantizer modules to PyTorch FakeQuantizer module and remove unused quantizer operators.

    :param model: Target model.
    :return: The modified NNCF network.
    """
    compression_module_type = ExtraCompressionModuleType.EXTERNAL_QUANTIZER
    if model.nncf.is_compression_module_registered(compression_module_type):
        external_quantizers = model.nncf.get_compression_modules_by_type(compression_module_type)
        for key in external_quantizers:
            if external_quantizers[key].is_enabled_quantization():
                external_quantizers[key] = convert_to_torch_fakequantizer(external_quantizers[key])

    for node in model.nncf.get_original_graph().get_all_nodes():
        if node.node_type in ["nncf_model_input", "nncf_model_output"]:
            continue

        nncf_module = model.nncf.get_containing_module(node.node_name)

        if hasattr(nncf_module, "pre_ops"):
            for key in list(nncf_module.pre_ops.keys()):
                op = nncf_module.get_pre_op(key)
                if isinstance(op.op, BaseQuantizer) and op.op.is_enabled_quantization():
                    if op.op.is_half_range or op.op.narrow_range:
                        # Half range and narrow_range require to clamp weights of module
                        # Note: Half range and narrow_range used only for weight.
                        input_low, input_high = op.op.get_input_low_input_high()

                        data = nncf_module.weight.data
                        data = torch.min(torch.max(data, input_low), input_high)
                        data = op.op.quantize(data, execute_traced_op_as_identity=False)
                        nncf_module.weight.data = data
                    op.op = convert_to_torch_fakequantizer(op.op)

        if hasattr(nncf_module, "post_ops"):
            for key in list(nncf_module.post_ops.keys()):
                op = nncf_module.get_post_ops(key)
                if isinstance(op.op, BaseQuantizer) and op.op.is_enabled_quantization():
                    op.op = convert_to_torch_fakequantizer(op.op)

    return model


def convert_to_torch_fakequantizer(nncf_quantizer: BaseQuantizer) -> FakeQuantize:
    """
    Convert BaseQuantizer module to FakeQuantize.

    :param quantizer: NNCF Quantizer module.
    :return: Instance of FakeQuantize similar to the input quantizer.
    """

    # Call set_ranges in case the basic parameters impacting levels had changed
    nncf_quantizer.set_levels()

    if nncf_quantizer.num_bits not in SUPPORTED_NUM_BITS_FOR_STRIP_MODEL:
        raise nncf.InternalError(
            "Converting nncf quantizer module to torch native only supports "
            f"for num_bits in {SUPPORTED_NUM_BITS_FOR_STRIP_MODEL}."
        )
    per_channel = nncf_quantizer.per_channel
    scale_shape = nncf_quantizer.scale_shape
    ch_axis = int(np.argmax(scale_shape))
    dtype = torch.qint8 if nncf_quantizer.level_low < 0 else torch.quint8

    if per_channel:
        observer = torch.ao.quantization.observer.PerChannelMinMaxObserver
    else:
        observer = torch.ao.quantization.observer.MinMaxObserver

    if isinstance(nncf_quantizer, SymmetricQuantizer):
        qscheme = torch.per_channel_symmetric if per_channel else torch.per_tensor_symmetric
    elif isinstance(nncf_quantizer, AsymmetricQuantizer):
        qscheme = torch.per_channel_affine if per_channel else torch.per_tensor_affine

    quant_min, quant_max, scale, zero_point = nncf_quantizer.get_parameters_for_torch_fq()

    fakequantizer = FakeQuantize(
        observer=observer,
        quant_max=quant_max,
        quant_min=quant_min,
        dtype=dtype,
        qscheme=qscheme,
        eps=nncf_quantizer.eps,
    )

    if not per_channel:
        scale = scale.squeeze()
        zero_point = zero_point.squeeze()

    fakequantizer.scale = scale
    fakequantizer.ch_axis = ch_axis
    fakequantizer.zero_point = zero_point

    # Disable observer to save parameters
    fakequantizer.disable_observer()

    return fakequantizer


def remove_disabled_quantizers(model: NNCFNetwork) -> NNCFNetwork:
    """
    Remove all unused quantizer operators from the model.

    :param model: Compressed model.
    :return: The modified NNCF network.
    """
    compression_module_type = ExtraCompressionModuleType.EXTERNAL_QUANTIZER
    if model.nncf.is_compression_module_registered(compression_module_type):
        external_quantizers = model.nncf.get_compression_modules_by_type(compression_module_type)
        for key in list(external_quantizers.keys()):
            op = external_quantizers[key]
            if isinstance(op, BaseQuantizer) and not op.is_enabled_quantization():
                external_quantizers.pop(key)

    if not model.nncf.replace_modules:
        return model

    for node in model.nncf.get_original_graph().get_all_nodes():
        if node.node_type in ["nncf_model_input", "nncf_model_output"]:
            continue

        nncf_module = model.nncf.get_containing_module(node.node_name)

        if hasattr(nncf_module, "pre_ops"):
            for key in list(nncf_module.pre_ops.keys()):
                op = nncf_module.get_pre_op(key)
                if isinstance(op, BaseQuantizer) and not op.is_enabled_quantization():
                    nncf_module.remove_pre_forward_operation(key)

        if hasattr(nncf_module, "post_ops"):
            for key in list(nncf_module.post_ops.keys()):
                op = nncf_module.post_ops(key)
                if isinstance(op, BaseQuantizer) and not op.is_enabled_quantization():
                    nncf_module.remove_post_forward_operation(key)

    return model


def strip_quantized_model(model: NNCFNetwork):
    """
    Returns the model with as much custom NNCF additions as possible removed
    while still preserving the functioning of the model object as a compressed model.

    :param model: Compressed model.
    :return: The modified NNCF network.
    """
    model = replace_quantizer_to_torch_native_module(model)
    model = remove_disabled_quantizers(model)
    return model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Iterable, List, Optional, Tuple

import torch

import nncf
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.experimental.tensor.tensor import Tensor
from nncf.parameters import CompressWeightsMode
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import compress_weight
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.graph.transformations.commands import PTSharedFnInsertionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import find_const_node_in_constant_subgraph
from nncf.torch.model_graph_manager import get_const_node
from nncf.torch.model_graph_manager import get_module_by_name
from nncf.torch.model_graph_manager import split_const_name
from nncf.torch.model_transformer import PTModelTransformer
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import WeightsDecompressor
from nncf.torch.tensor_statistics.collectors import get_raw_stat_collector


class PTWeightCompressionAlgoBackend(WeightCompressionAlgoBackend):
    TARGET_TYPE_TO_PT_INS_TYPE_MAP = {
        TargetType.PRE_LAYER_OPERATION: TargetType.OPERATOR_PRE_HOOK,
        TargetType.POST_LAYER_OPERATION: TargetType.OPERATOR_POST_HOOK,
    }
    MATMUL_METATYPES = [om.PTLinearMetatype, om.PTMatMulMetatype, om.PTAddmmMetatype]
    EMBEDDING_METATYPES = [om.PTEmbeddingMetatype]
    CONVOLUTION_METATYPES = [
        om.PTConv1dMetatype,
        om.PTConv2dMetatype,
        om.PTConv3dMetatype,
        om.PTDepthwiseConv1dSubtype,
        om.PTDepthwiseConv2dSubtype,
        om.PTDepthwiseConv3dSubtype,
        om.PTConvTranspose1dMetatype,
        om.PTConvTranspose2dMetatype,
        om.PTConvTranspose3dMetatype,
    ]

    @property
    def matmul_metatypes(self) -> List[OperatorMetatype]:
        return PTWeightCompressionAlgoBackend.MATMUL_METATYPES

    @property
    def embedding_metatypes(self) -> List[OperatorMetatype]:
        return PTWeightCompressionAlgoBackend.EMBEDDING_METATYPES

    @property
    def convolution_metatypes(self) -> List[OperatorMetatype]:
        return PTWeightCompressionAlgoBackend.CONVOLUTION_METATYPES

    @staticmethod
    def is_node_with_weights(node: NNCFNode, graph: NNCFGraph) -> bool:
        if (
            node.metatype not in PTWeightCompressionAlgoBackend.MATMUL_METATYPES
            and node.metatype not in PTWeightCompressionAlgoBackend.EMBEDDING_METATYPES
            and node.metatype not in PTWeightCompressionAlgoBackend.CONVOLUTION_METATYPES
        ):
            return False
        for prev_node in graph.get_previous_nodes(node):
            edge = graph.get_edge(prev_node, node)
            if edge.input_port_id not in node.metatype.weight_port_ids:
                continue
            weight_node = find_const_node_in_constant_subgraph(prev_node, graph)
            if weight_node is not None:
                return True
        return False

    @staticmethod
    def get_weight_names_and_port_ids(node: NNCFNode, graph: NNCFGraph) -> List[Tuple[str, int]]:
        weight_port_ids = []
        for prev_node in graph.get_previous_nodes(node):
            weight_node = find_const_node_in_constant_subgraph(prev_node, graph)
            if weight_node is None:
                continue
            edge = graph.get_edge(prev_node, node)
            if edge.input_port_id in node.metatype.weight_port_ids:
                weight_port_ids.append((weight_node.layer_attributes.name, edge.input_port_id))
        return weight_port_ids

    @staticmethod
    def get_reduction_axes(node_with_weight: NNCFNode, weight_port_id: int, graph: NNCFGraph) -> Optional[Tuple[int]]:
        weight_node = get_const_node(node_with_weight, weight_port_id, graph)

        ndims = len(weight_node.layer_attributes.shape)
        reduction_axes = None
        if node_with_weight.metatype == om.PTEmbeddingMetatype:
            reduction_axes = [1]
        elif node_with_weight.metatype == om.PTLinearMetatype:
            reduction_axes = [ndims - 1]
        elif node_with_weight.metatype == om.PTMatMulMetatype:
            if weight_port_id == 0:
                reduction_axes = [ndims - 1]
            elif weight_port_id == 1:
                reduction_axes = [max(0, ndims - 2)]
        elif node_with_weight.metatype == om.PTAddmmMetatype:
            if weight_port_id == 1:
                reduction_axes = [ndims - 1]
            elif weight_port_id == 2:
                reduction_axes = [max(0, ndims - 2)]
        elif node_with_weight.metatype in PTWeightCompressionAlgoBackend.CONVOLUTION_METATYPES:
            channel_idx = (
                1
                if node_with_weight.metatype
                in [om.PTConvTranspose1dMetatype, om.PTConvTranspose2dMetatype, om.PTConvTranspose3dMetatype]
                else 0
            )
            reduction_axes = [i for i in range(ndims) if i != channel_idx]
        return tuple(reduction_axes)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        if NNCFGraphNodeType.INPUT_NODE in target_node_name or target_type == TargetType.POST_LAYER_OPERATION:
            port_id = None
        if target_type in PTWeightCompressionAlgoBackend.TARGET_TYPE_TO_PT_INS_TYPE_MAP:
            target_type = PTWeightCompressionAlgoBackend.TARGET_TYPE_TO_PT_INS_TYPE_MAP[target_type]
        return PTTargetPoint(target_type, target_node_name, input_port_id=port_id)

    @staticmethod
    def raw_statistic_collector(num_samples: Optional[int] = None) -> TensorCollector:
        return get_raw_stat_collector(num_samples)

    @staticmethod
    def get_activation_port_id(node: NNCFNode, graph: NNCFGraph) -> int:
        activation_ports = []
        for prev_node in graph.get_previous_nodes(node):
            if prev_node.metatype in CONST_NOOP_METATYPES:
                continue
            edge = graph.get_edge(prev_node, node)
            activation_ports.append(edge.input_port_id)
        assert len(activation_ports) == 1
        return activation_ports[0]

    def get_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: torch.nn.Module, graph: NNCFGraph
    ) -> Tensor:
        weight_node = get_const_node(node_with_weight, weight_port_id, graph)
        weight_name = weight_node.layer_attributes.name
        module_name, weight_attr_name = split_const_name(weight_name)
        module = get_module_by_name(module_name, model)
        weight = getattr(module, weight_attr_name)
        if weight is None or not isinstance(weight, torch.nn.Parameter):
            raise nncf.InternalError(f"Could not find a torch.nn.Parameter in the model by name {weight_name}.")

        return Tensor(weight)

    def set_weight(
        self, node_with_weight: NNCFNode, weight_port_id: int, model: torch.nn.Module, graph: NNCFGraph, weight: Tensor
    ):
        pass

    def transform_model(
        self,
        model: NNCFNetwork,
        graph: NNCFGraph,
        weight_compression_parameters: Iterable[WeightCompressionParameters],
        precomputed_scales: Dict[str, Tensor] = None,
        precomputed_zero_points: Dict[str, Tensor] = None,
    ) -> NNCFNetwork:
        transformation_layout = TransformationLayout()

        for wc_params in weight_compression_parameters:
            compression_config = wc_params.compression_config
            if compression_config.mode not in [
                CompressWeightsMode.INT8_ASYM,
                CompressWeightsMode.INT8_SYM,
                CompressWeightsMode.INT8,
            ]:
                raise ValueError(f"{compression_config.mode.value} is not supported.")

            weight_node = get_const_node(wc_params.node_with_weight, wc_params.weight_port_id, graph)
            weight_name = weight_node.layer_attributes.name
            module_name, weight_attr_name = split_const_name(weight_name)
            module = get_module_by_name(module_name, model)
            weight = getattr(module, weight_attr_name)
            if weight is None or not isinstance(weight, torch.nn.Parameter):
                raise nncf.InternalError(f"Could not find a torch.nn.Parameter in the model by name {weight_name}.")

            # calculates compressed weights and decompression parameters
            compressed_weight = compress_weight(
                Tensor(weight),
                wc_params.reduction_axes,
                compression_config,
                None if precomputed_scales is None else precomputed_scales.get(wc_params.weight_name),
                None if precomputed_zero_points is None else precomputed_zero_points.get(wc_params.weight_name),
            )
            compressed_weight.scale = compressed_weight.scale.astype(dtype=TensorDataType.float16)

            # pack compressed tensor
            packed_tensor = compressed_weight.tensor.astype(TensorDataType.uint8)

            # sets compressed tensor
            compressed_parameter = torch.nn.Parameter(packed_tensor.data, requires_grad=False)
            setattr(module, weight_attr_name, compressed_parameter)

            consumer_nodes = graph.get_next_nodes(weight_node)
            if len(consumer_nodes) > 1:
                for c_node in consumer_nodes:
                    c_module = model.nncf.get_module_by_scope(Scope.from_str(c_node.layer_name))
                    for name, param in c_module.named_parameters(recurse=False, remove_duplicate=False):
                        if id(param) == id(weight):
                            setattr(c_module, name, compressed_parameter)

            # pack zero point tensor
            packed_zero_point = compressed_weight.zero_point.astype(TensorDataType.uint8)

            # creates weight decompressor
            decompressor = WeightsDecompressor(
                compressed_weight.scale.data, packed_zero_point.data, result_dtype=weight.dtype
            )

            # registry weight decompression module in the model
            decompressor_name = f"weights_decompressor_{weight_node.node_name.replace('.', '_')}"

            # inserts the weight decompressor into the model as the post hook on the model weight
            transformation_layout.register(
                PTSharedFnInsertionCommand(
                    [PTTargetPoint(TargetType.OPERATOR_POST_HOOK, target_node_name=weight_node.node_name)],
                    decompressor,
                    decompressor_name,
                )
            )

        # apply transformations
        transformed_model = PTModelTransformer(model).transform(transformation_layout)

        return transformed_model



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noqa
from nncf.experimental.torch.nas.bootstrapNAS.elasticity import elasticity_builder as elasticity_algo
from nncf.experimental.torch.nas.bootstrapNAS.search.search import BaseSearchAlgorithm
from nncf.experimental.torch.nas.bootstrapNAS.training import progressive_shrinking_builder as ps_algo
from nncf.experimental.torch.nas.bootstrapNAS.training.training_algorithm import EpochBasedTrainingAlgorithm



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import numpy as np
import onnx

from nncf.common.factory import TModel
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.tensor_statistics.aggregator import StatisticsAggregator
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.onnx.graph.node_utils import get_input_edge
from nncf.onnx.graph.node_utils import get_input_edges_mapping
from nncf.onnx.graph.onnx_helper import get_name_to_node_map
from nncf.onnx.graph.transformations.commands import ONNXOutputInsertionCommand
from nncf.onnx.tensor import ONNXNNCFTensor


class ONNXStatisticsAggregator(StatisticsAggregator):
    def collect_statistics(self, model: onnx.ModelProto, graph: NNCFGraph) -> None:
        self.input_edges_mapping = get_input_edges_mapping(graph)
        self.node_mapping = get_name_to_node_map(model)
        self._registered_weights = set()
        super().collect_statistics(model, graph)

    def _register_statistics(
        self, outputs: Dict[str, ONNXNNCFTensor], statistic_points: StatisticPointsContainer
    ) -> None:
        for _, statistic_point, tensor_collector in statistic_points.get_tensor_collectors():
            target_point = statistic_point.target_point
            port_id = target_point.port_id

            if target_point.target_node_name in self.input_edges_mapping:  # Input case
                edge_name = get_input_edge(
                    target_point.target_node_name,
                    self.input_edges_mapping,
                    self.node_mapping,
                )
            elif target_point.type == TargetType.POST_LAYER_OPERATION:
                node = self.node_mapping[target_point.target_node_name]
                edge_name = node.output[port_id]
            elif target_point.type in [TargetType.PRE_LAYER_OPERATION, TargetType.OPERATION_WITH_WEIGHTS]:
                node = self.node_mapping[target_point.target_node_name]
                edge_name = node.input[port_id]
            else:
                RuntimeError(f"Unsupported target point type for statistic aggregator: {target_point.type}")

            input_info = []
            for reducer in tensor_collector.reducers:
                input_info.append((hash(reducer), [edge_name]))

            target_inputs = TensorCollector.get_tensor_collector_inputs(outputs, input_info)
            tensor_collector.register_inputs(target_inputs)

    def _get_transformation_layout_extra_outputs(
        self, statistic_points: StatisticPointsContainer
    ) -> TransformationLayout:
        transformation_layout = TransformationLayout()
        transformation_commands = []
        for _statistic_points in statistic_points.values():
            for _statistic_point in _statistic_points:
                transformation_commands.append(
                    ONNXOutputInsertionCommand(_statistic_point.target_point, self.input_edges_mapping)
                )
        for transformation_command in transformation_commands:
            transformation_layout.register(transformation_command)

        return transformation_layout

    @staticmethod
    def _get_merged_statistic_points(
        statistic_points: StatisticPointsContainer, model: TModel, graph: NNCFGraph
    ) -> StatisticPointsContainer:
        # TODO: mirgate to experimental statistic collector and use common merging algorithm
        return statistic_points

    @staticmethod
    def _process_outputs(outputs: Dict[str, np.ndarray]) -> Dict[str, ONNXNNCFTensor]:
        return {n: ONNXNNCFTensor(v) for n, v in outputs.items()}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from abc import abstractmethod
from typing import Any, List, Optional, Tuple

from nncf.api.compression import TModel


class Exporter(ABC):
    """
    This is the class from which all framework-specific exporters inherit.

    An exporter is an object which provides export of the compressed model
    for deployment.
    """

    def __init__(
        self,
        model: TModel,
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        model_args: Optional[Tuple[Any, ...]] = None,
    ):
        """
        Initializes an exporter.

        :param model: The model to be exported.
        :param input_names: Names to be assigned to the input tensors of the model.
        :param output_names: Names to be assigned to the output tensors of the model.
        :param model_args: Tuple of additional positional and keyword arguments
            which are required for the model's forward during export. Should be
            specified in the following format:
                - (a, b, {'x': None, 'y': y}) for positional and keyword arguments.
                - (a, b, {}) for positional arguments only.
                - ({'x': None, 'y': y},) for keyword arguments only.
        """
        self._model = model
        self._input_names = input_names
        self._output_names = output_names
        self._model_args = model_args

    @abstractmethod
    def export_model(self, save_path: str, save_format: Optional[str] = None) -> None:
        """
        Exports the compressed model to the specified format.

        :param save_path: The path where the model will be saved.
        :param save_format: Saving format. The default format will
            be used if `save_format` is not specified.
        """



# Copyright (c) 2023 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial

from nncf.common.graph.patterns import GraphPattern
from nncf.common.utils.registry import Registry

AWQ_PATTERNS = Registry("awq")


@AWQ_PATTERNS.register("MatMul_Mul_MatMul")
def create_matmul_mul_matmul(matmul_metatype, multiply_metatype) -> GraphPattern:
    pattern = GraphPattern()
    linear_node_1 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "LINEAR", GraphPattern.METATYPE_ATTR: matmul_metatype})
    mul_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MULTIPLY", GraphPattern.METATYPE_ATTR: multiply_metatype})
    linear_node_2 = pattern.add_node(**{GraphPattern.LABEL_ATTR: "LINEAR", GraphPattern.METATYPE_ATTR: matmul_metatype})

    pattern.add_edge(linear_node_1, mul_node)
    pattern.add_edge(mul_node, linear_node_2)
    return pattern


def get_awq_patterns(matmul_metatype, multiply_metatype):
    res = Registry("awq")
    for k, v in AWQ_PATTERNS.registry_dict.items():
        res.registry_dict[k] = partial(v, matmul_metatype, multiply_metatype)

    return res.registry_dict



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph import NNCFNodeName


class PrunedLayerInfoBase:
    def __init__(self, node_name: NNCFNodeName, node_id: int, is_depthwise: bool):
        self.node_name = node_name
        self.nncf_node_id = node_id
        self.is_depthwise = is_depthwise



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from nncf.common.sparsity.collector import BaseSparseModelStatisticsCollector
from nncf.common.sparsity.collector import WeightDescription
from nncf.torch.layer_utils import COMPRESSION_MODULES
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.sparsity.base_algo import SparseModuleInfo


class PTSparseModelStatisticsCollector(BaseSparseModelStatisticsCollector):
    """
    Collects statistics for the sparse NNCFNetwork.
    """

    def __init__(
        self, model: NNCFNetwork, sparse_modules_info: List[SparseModuleInfo], supports_sparse_bias: bool = False
    ):
        """
        Initializes statistics collector of the sparse tf.keras.Model.

        :param model: Sparse model.
        :param sparse_modules_info: List of `SparseModuleInfo`.
        """
        self._model = model
        self._sparse_modules_info = sparse_modules_info
        self._supports_sparse_bias = supports_sparse_bias

    def _collect_weights_descriptions(self) -> List[WeightDescription]:
        weights_descriptions = []
        processed_modules = []

        for minfo in self._sparse_modules_info:
            sparse_weight = minfo.operand.apply_binary_mask(minfo.module.weight)

            weights_descriptions.append(
                WeightDescription(
                    minfo.module_node_name,
                    list(sparse_weight.shape),
                    sparse_weight.count_nonzero().item(),
                    is_sparse=True,
                )
            )

            if hasattr(minfo.module, "bias") and minfo.module.bias is not None:
                bias = minfo.module.bias
                name = f"{minfo.module_node_name}/bias"
                if self._supports_sparse_bias:
                    sparse_bias = minfo.operand.apply_binary_mask(bias, is_bias=True)  # TODO(yujie): breaking changes
                    weights_descriptions.append(
                        WeightDescription(
                            name, list(sparse_bias.shape), sparse_bias.count_nonzero().item(), is_sparse=True
                        )
                    )
                else:
                    weights_descriptions.append(
                        WeightDescription(name, list(bias.shape), bias.count_nonzero().item(), is_sparse=False)
                    )

            processed_modules.append(minfo.module)

        compression_types = tuple(COMPRESSION_MODULES.registry_dict.values())
        modules_to_process = {k: v for k, v in self._model.named_modules() if not k.startswith("_nncf")}
        for module_name, module in modules_to_process.items():
            if isinstance(module, compression_types) or module in processed_modules:
                continue

            for param_name, param in module.named_parameters(recurse=False):
                name = f"{module_name}/{param_name}"
                weights_descriptions.append(
                    WeightDescription(name, list(param.shape), param.count_nonzero().item(), is_sparse=False)
                )

        return weights_descriptions



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for user-defined data usage during the compression algorithm initialization process."""
from abc import ABC
from abc import abstractmethod

from nncf.common.utils.api_marker import api


@api()
class NNCFDataLoader(ABC):
    """
    Wraps a custom data source.
    """

    @property
    @abstractmethod
    def batch_size(self) -> int:
        """
        Returns the number of elements return per iteration.

        :return: A number of elements return per iteration.
        """

    @abstractmethod
    def __iter__(self):
        """
        Creates an iterator for the elements of a custom data source.
        The returned iterator implements the Python Iterator protocol.

        :return: An iterator for the elements of a custom data source.
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import List, Union

from nncf.common.tensor import DeviceType
from nncf.common.tensor import NNCFTensor


class NNCFPruningBaseTensorProcessor:
    """
    An interface of the processing methods for NNCFTensors.
    """

    @classmethod
    @abstractmethod
    def concatenate(cls, tensors: List[NNCFTensor], axis: int) -> NNCFTensor:
        """
        Join a list of NNCFTensors along an existing axis.

        :param tensors: List of NNCFTensors.
        :param axis: The axis, along which the tensors will be joined.
        :returns: The concatenated List of the tensors.
        """

    @classmethod
    @abstractmethod
    def ones(cls, shape: Union[int, List[int]], device: DeviceType) -> NNCFTensor:
        """
        Return a new float tensor of given shape, filled with ones.

        :param shape: Shape of the new tensor.
        :param device: Device to put created tensor in.
        :returns: Float tensor of ones with the given shape.
        """

    @classmethod
    @abstractmethod
    def assert_allclose(cls, tensors: List[NNCFTensor]) -> None:
        """
        Raises an AssertionError if any two tensors are not equal.

        :param tensors: List of tensors to check pairwise equality.
        """

    @classmethod
    @abstractmethod
    def repeat(cls, tensor: NNCFTensor, repeats: int) -> NNCFTensor:
        """
        Successively repeat each element of given NNCFTesnor.

        :param tensor: Given NNCFTensor.
        :param repeats: The number of repetitions for each element.
        :return: NNCFTensor with repited elements.
        """

    @classmethod
    @abstractmethod
    def elementwise_mask_propagation(cls, input_masks: List[NNCFTensor]) -> NNCFTensor:
        """
        Assemble output mask for elementwise pruning operation from given input masks.
        Raises an AssertionError if input masks are not pairwise equal.

        :param input_masks: Given input masks.
        :return: Elementwise pruning operation output mask.
        """

    @classmethod
    @abstractmethod
    def split(cls, tensor: NNCFTensor, output_shapes: List[int]) -> List[NNCFTensor]:
        """
        Split/chunk NNCFTensor into chunks along an exsiting dimension.

        :param tensor: Given NNCFTensor.
        :param output_shapes: Given shapes of the output masks
        :returns: The list of NNCFTensor which is split
        """



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
from dataclasses import dataclass
from dataclasses import field
from typing import Dict, List, NamedTuple, Set, TypeVar

from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.quantization.passes import transform_to_inference_graph

TModel = TypeVar("TModel")


class NodeOutputPort(NamedTuple):
    """
    Represents the output port of a node in the graph.

    :param node_name: The name of the node.
    :param output_port: The output port index.
    """

    node_name: str
    output_port: int


@dataclass
class LayerwiseStep:
    """
    Represents a single step in the layer-wise processing of the graph.

    :param target_node_map: Mapping of target nodes to their inputs/outputs ports
        and corresponding output ports in the subgraph.
    :param subgraph_inputs: List of input ports for the subgraph.
    :param subgraph_outputs: List of output ports for the subgraph.
    """

    target_node_map: Dict[NNCFNode, Dict[int, NodeOutputPort]]
    subgraph_inputs: List[NodeOutputPort]
    subgraph_outputs: List[NodeOutputPort]


@dataclass
class SimplePath:
    """
    Represents a simple path in the graph with input and output nodes.

    :param input_nodes: Set of input nodes.
    :param output_nodes: Set of output nodes.
    :param inputs: Set of input ports.
    """

    input_nodes: Set[NNCFNode]
    output_nodes: Set[NNCFNode]
    inputs: Set[NodeOutputPort] = field(default_factory=set)


class LayerwiseScheduler:
    """
    Scheduler that determines the order of layer-wise steps for processing a graph based on a given strategy.

    The scheduling algorithm works as follows:
    1. Initialize input nodes and create a copy of the graph for inference.
    2. Perform a topological traversal of the graph starting from the input nodes.
    3. For each node, if it is a target node and has not been processed, add the path from inputs to this node
       to the paths list.
    4. If `add_additional_outputs` is True, add all output nodes which was partial visited to remove inputs
       to target nodes.
    5. Merge paths that have overlapping input sets into single paths to minimize model extraction time.
    6. Create layer-wise steps by mapping target nodes to their input/output ports and corresponding output ports
       in the subgraph.
    7. Repeat the process until all target nodes have been processed.
    """

    def __init__(self, add_additional_outputs: bool = False):
        """
        :param strategy: The strategy to use for scheduling.
        """
        self.add_additional_outputs = add_additional_outputs

    def schedule(
        self, graph: NNCFGraph, target_nodes: List[NNCFNode], collect_inputs: bool = True
    ) -> List[LayerwiseStep]:
        """
        Schedules the execution of the graph in a layer-wise manner.

        :param graph: The graph to be scheduled.
        :param target_nodes: The nodes that are targeted for layer-wise processing.
        :param collect_inputs: If True (default) collects inputs for target nodes
            oterwise collects outputs of target nodes.
        :return: The scheduled steps for layer-wise processing.
        """
        # Initialize input nodes and create a copy of the graph for inference
        input_nodes = graph.get_input_nodes()
        inference_graph = transform_to_inference_graph(deepcopy(graph), input_nodes, [], [])

        steps = []
        visited_map = {node: False for node in inference_graph.get_all_nodes()}
        while not all([visited_map[node] for node in target_nodes]):
            input_nodes_map = {node: {node} for node in input_nodes}
            zero_indegree = input_nodes
            indegree_map = {}
            innode_map = {}
            paths = []

            # topological traversal of a graph starting from input nodes
            while zero_indegree:
                this_generation = zero_indegree
                zero_indegree = []
                for node in this_generation:
                    for next in inference_graph.get_next_nodes(node):
                        if next in input_nodes:
                            continue

                        if next not in input_nodes_map:
                            input_nodes_map[next] = set()
                        input_nodes_map[next] |= input_nodes_map[node]

                        if next not in indegree_map:
                            indegree_map[next] = len(inference_graph.get_previous_nodes(next))
                        indegree_map[next] -= 1

                        if indegree_map[next] == 0:
                            # if node is a target node and is not processed yet then
                            # the path from inputs to node is added to paths
                            if next not in target_nodes:
                                zero_indegree.append(next)
                            elif not visited_map[next]:
                                paths.append(SimplePath(input_nodes=input_nodes_map[next], output_nodes={next}))
                            visited_map[next] = True
                            del indegree_map[next]
                        else:
                            if next not in innode_map:
                                innode_map[next] = set()
                            innode_map[next].add(node)

            # if add_additional_outputs = True then add additional outputs to minimize computation cost,
            # otherwise reuse existing inputs.
            reuse_input_nodes = set()
            if self.add_additional_outputs:
                for node in indegree_map:
                    if not visited_map[node]:
                        paths.append(SimplePath(input_nodes_map[node], innode_map[node]))
            else:
                for node in indegree_map:
                    if not visited_map[node]:
                        reuse_input_nodes |= input_nodes_map[node]

            # fill input ports
            inputs_map = {}
            for node in input_nodes:
                inputs_map[node] = set()
                if collect_inputs and node in target_nodes:
                    for edge in inference_graph.get_input_edges(node):
                        inputs_map[node].add(NodeOutputPort(edge.from_node.node_name, edge.output_port_id))
                else:
                    for edge in inference_graph.get_output_edges(node):
                        inputs_map[node].add(NodeOutputPort(node.node_name, edge.output_port_id))

            for path in paths:
                for input_node in path.input_nodes:
                    path.inputs |= inputs_map[input_node]

            # merge paths to minimize model extraction time
            paths = self._merge_paths(paths)

            # create layerwise step by paths
            old_input_nodes = set()
            new_input_nodes = set()
            for p in paths:
                target_output_nodes = set()
                additional_output_nodes = set()
                for output_node in p.output_nodes:
                    if output_node in target_nodes:
                        target_output_nodes.add(output_node)
                    elif output_node in p.input_nodes:
                        reuse_input_nodes.add(output_node)
                    else:
                        # filter additional output nodes
                        for prev_node in inference_graph.get_previous_nodes(output_node):
                            if prev_node not in p.output_nodes:
                                additional_output_nodes.add(output_node)
                                break
                if not target_output_nodes:
                    continue

                old_input_nodes |= p.input_nodes
                new_input_nodes |= target_output_nodes | additional_output_nodes
                subgraph_inputs = list(p.inputs)
                step_target_nodes = {}
                subgraph_outputs = []
                for node in target_output_nodes:
                    target_edge = {}
                    if collect_inputs:
                        for edge in inference_graph.get_input_edges(node):
                            output_id = NodeOutputPort(edge.from_node.node_name, edge.output_port_id)
                            target_edge[edge.input_port_id] = output_id
                            if output_id not in subgraph_outputs:
                                subgraph_outputs.append(output_id)
                    else:
                        for edge in inference_graph.get_output_edges(node):
                            output_id = NodeOutputPort(edge.from_node.node_name, edge.output_port_id)
                            target_edge[edge.output_port_id] = output_id
                            if output_id not in subgraph_outputs:
                                subgraph_outputs.append(output_id)
                    step_target_nodes[graph.get_node_by_name(node.node_name)] = target_edge
                for node in additional_output_nodes:
                    for edge in inference_graph.get_output_edges(node):
                        output_id = NodeOutputPort(edge.from_node.node_name, edge.output_port_id)
                        if output_id not in subgraph_outputs:
                            subgraph_outputs.append(output_id)
                steps.append(LayerwiseStep(step_target_nodes, subgraph_inputs, subgraph_outputs))

            # prepare input nodes for next steps
            if self.add_additional_outputs:
                reuse_input_nodes |= set(input_nodes) - old_input_nodes
            input_nodes = list(new_input_nodes | reuse_input_nodes)

        return steps

    def _merge_paths(self, paths: List[SimplePath]) -> List[SimplePath]:
        """
        Merges paths that have overlapping input sets into single paths.

        :param paths: List of paths to merge.
        :returns: The merged list of paths.
        """
        result = []
        for p in paths:
            for s in result:
                if not s.inputs.isdisjoint(p.inputs):
                    s.inputs |= p.inputs
                    s.input_nodes |= p.input_nodes
                    s.output_nodes |= p.output_nodes
                    break
            else:
                result.append(p)
        if len(paths) == len(result):
            return result
        return self._merge_paths(result)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

from nncf.common.quantization.quantizer_propagation.structs import QuantizationTrait
from nncf.torch.graph import operator_metatypes
from nncf.torch.graph.operator_metatypes import OPERATORS_WITH_WEIGHTS_METATYPES
from nncf.torch.graph.operator_metatypes import PTOperatorMetatype

# If a metatype is not in this list, then it is considered to be QuantizationTrait.NON_QUANTIZABLE.

DEFAULT_PT_QUANT_TRAIT_TO_OP_DICT: Dict[QuantizationTrait, List[PTOperatorMetatype]] = {
    QuantizationTrait.INPUTS_QUANTIZABLE: [
        operator_metatypes.PTConv2dMetatype,
        operator_metatypes.PTModuleConv2dMetatype,
        operator_metatypes.PTConv3dMetatype,
        operator_metatypes.PTModuleConv3dMetatype,
        operator_metatypes.PTConvTranspose2dMetatype,
        operator_metatypes.PTModuleConvTranspose2dMetatype,
        operator_metatypes.PTConvTranspose3dMetatype,
        operator_metatypes.PTModuleConvTranspose3dMetatype,
        operator_metatypes.PTDepthwiseConv2dSubtype,
        operator_metatypes.PTDepthwiseConv3dSubtype,
        operator_metatypes.PTModuleDepthwiseConv2dSubtype,
        operator_metatypes.PTModuleDepthwiseConv3dSubtype,
        operator_metatypes.PTLinearMetatype,
        operator_metatypes.PTModuleLinearMetatype,
        operator_metatypes.PTLayerNormMetatype,
        operator_metatypes.PTModuleLayerNormMetatype,
        operator_metatypes.PTAddMetatype,
        operator_metatypes.PTMulMetatype,
        operator_metatypes.PTDivMetatype,
        operator_metatypes.PTMatMulMetatype,
        operator_metatypes.PTMeanMetatype,
        operator_metatypes.PTRoundMetatype,
        operator_metatypes.PTPixelShuffleMetatype,
        operator_metatypes.PTBatchNormMetatype,
        operator_metatypes.PTModuleBatchNormMetatype,
        operator_metatypes.PTAvgPool2dMetatype,
        operator_metatypes.PTAvgPool3dMetatype,
        # 1. Single input activations except Relu and PRelu could not be
        # executed in INT8 precision by the OpenVINO runtime.
        # List of supported operations for INT8 execution:
        # https://docs.openvino.ai/2023.1/openvino_docs_OV_UG_lpt.html#input-model-requirements
        # 2. In case an activation from Torch is fused to
        # a specific OpenVINO operation in runtime, it is better to not quantize
        # this actictivation to keep specific operations fusing.
        # operator_metatypes.PTHardTanhMetatype,
        # operator_metatypes.PTHardSwishMetatype,
        # operator_metatypes.PTHardSigmoidMetatype,
        # operator_metatypes.PTTanhMetatype,
        # operator_metatypes.PTELUMetatype,
        # operator_metatypes.PTLeakyRELUMetatype,
        # operator_metatypes.PTGELUMetatype,
        # operator_metatypes.PTErfMetatype,
        # PTPRELUMetatype is not considered to be QUANTIZATION_AGNOSTIC, because:
        # 1. Runtime doesn't provide performance benefits by quantizing the stand-alone RELU's (ticket: 59548)
        # 2. It's frequently better for the end accuracy to have quantizers set up after the RELU
        # so that the input distribution to the quantizer is non-negative
        # and we can therefore have better quantization resolution while preserving the original dynamic range
        # operator_metatypes.PTPRELUMetatype,
    ],
    QuantizationTrait.QUANTIZATION_AGNOSTIC: [
        operator_metatypes.PTThresholdMetatype,
        operator_metatypes.PTDropoutMetatype,
        operator_metatypes.PTPadMetatype,
        operator_metatypes.PTMaxMetatype,
        operator_metatypes.PTMinMetatype,
        operator_metatypes.PTTransposeMetatype,
        operator_metatypes.PTGatherMetatype,
        operator_metatypes.PTScatterMetatype,
        operator_metatypes.PTReshapeMetatype,
        operator_metatypes.PTSqueezeMetatype,
        operator_metatypes.PTSplitMetatype,
        operator_metatypes.PTExpandMetatype,
        operator_metatypes.PTMaxPool1dMetatype,
        operator_metatypes.PTMaxPool2dMetatype,
        operator_metatypes.PTMaxPool3dMetatype,
        operator_metatypes.PTMaxUnpool1dMetatype,
        operator_metatypes.PTMaxUnpool2dMetatype,
        operator_metatypes.PTMaxUnpool3dMetatype,
        operator_metatypes.PTRepeatMetatype,
        operator_metatypes.PTNoopMetatype,
        # PTRELUMetatype is not considered to be QUANTIZATION_AGNOSTIC, because:
        # 1. Runtime doesn't provide performance benefits by quantizing the stand-alone RELU's (ticket: 59548)
        # 2. It's frequently better for the end accuracy to have quantizers set up after the RELU
        # so that the input distribution to the quantizer is non-negative
        # and we can therefore have better quantization resolution while preserving the original dynamic range
    ],
    QuantizationTrait.CONCAT: [operator_metatypes.PTCatMetatype],
    QuantizationTrait.OUTPUT_QUANTIZATION_AS_WEIGHTS: [
        operator_metatypes.PTEmbeddingMetatype,
        operator_metatypes.PTModuleEmbeddingMetatype,
        operator_metatypes.PTEmbeddingBagMetatype,
        operator_metatypes.PTModuleEmbeddingBagMetatype,
    ],
}


QUANTIZATION_LAYER_METATYPES: List[PTOperatorMetatype] = OPERATORS_WITH_WEIGHTS_METATYPES



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import islice
from typing import List, Optional, Tuple

import openvino.runtime as ov

from nncf import Dataset
from nncf.common import factory
from nncf.common.engine import Engine
from nncf.common.factory import NNCFGraphFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.openvino.graph.metatypes.openvino_metatypes import OVIfMetatype
from nncf.openvino.graph.model_utils import remove_friendly_name_duplicates
from nncf.openvino.graph.node_utils import get_number_if_op
from nncf.openvino.graph.transformations.commands import OVExtractIfBodyCommand
from nncf.openvino.graph.transformations.commands import OVOutputInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.graph.transformations.commands import OVUpdateIfBodyCommand
from nncf.quantization.algorithms.algorithm import Algorithm


def _make_dataset_for_if_bodies(
    engine: Engine,
    calibration_dataset: Dataset,
    if_cond_input_name: str,
    then_model_input_names: List[str],
    else_model_input_names: List[str],
    subset_size: int,
) -> Tuple[Dataset, Dataset]:
    """
    Returns dataset for a then and else bodies of If node.

    :param engine: Engine to infer parent model to obtain dataitems for a child dataset.
    :param calibration_dataset: Dataset to infer parent model.
    :param if_cond_input_name: Input name of If node condition.
    :param then_model_input_names: Names of inputs for then body
    (should be in the order of passing them to a model).
    :param else_model_input_names: Names of inputs for else body
    (should be in the order of passing them to a model).
    :param subset_size: The size of calibration_dataset.
    :return Dataset: Dataset for child model.
    """

    then_dataset, else_dataset = [], []
    calibration_dataset_size = (
        min(subset_size, calibration_dataset.get_length())
        if calibration_dataset.get_length() is not None
        else subset_size
    )
    for input_data in track(
        islice(calibration_dataset.get_inference_data(), calibration_dataset_size),
        total=calibration_dataset_size,
        description="Collecting the dataset for then and else bodies:",
    ):
        data_item = []
        results = engine.infer(input_data)
        if results[if_cond_input_name]:
            for name in then_model_input_names:
                data_item.append(results[name])
            then_dataset.append(data_item)
        else:
            for name in else_model_input_names:
                data_item.append(results[name])
            else_dataset.append(data_item)
    nncf_logger.info(f"The length of dataset for then body is {len(then_dataset)}, else body is {len(else_dataset)}.")
    return Dataset(then_dataset), Dataset(else_dataset)


def _extract_if_body(model_transformer: ModelTransformer, if_node: NNCFNode, if_body_condition: bool) -> ov.Model:
    """
    Returns if body of If node based on a value of if_body_condition.

    :param model_transformer: ModelTransformer instance.
    :param if_node: If node.
    :param if_submodel_condition: If True returns then body of If node, otherwise - else body.
    :return: If body.
    """
    transformation_layout = TransformationLayout()
    command = OVBackend.create_extract_if_body_command(if_node, if_body_condition)
    transformation_layout.register(command)
    return model_transformer.transform(transformation_layout)


def _update_if_body(
    model_transformer: ModelTransformer, if_node: NNCFNode, if_body_condition: bool, body: ov.Model
) -> ov.Model:
    """
    Update body of If node, based on if_body_condition.

    :param model_transformer: ModelTransformer instance.
    :param if_node: If node.
    :param if_body_condition: Condition of If node body.
    :param body: New body.
    :return: Updated model with a new body of If node.
    """
    transformation_layout = TransformationLayout()
    command = OVBackend.create_update_body_command(if_node, if_body_condition, body)
    transformation_layout.register(command)
    return model_transformer.transform(transformation_layout)


def _add_outputs_before_if_node(model_transformer: ModelTransformer, model: ov.Model, if_node: NNCFNode) -> ov.Model:
    """
    Inserts extra outputs on If node inputs.

    :param model_transformer: ModelTransformer instance.
    :param model: Model instance.
    :param if_node: If node.
    :return: Model with extra outputs before If node.
    """
    transformation_layout = TransformationLayout()
    for command in OVBackend.create_output_insertion_commands_if_node(model, if_node):
        transformation_layout.register(command)
    return model_transformer.transform(transformation_layout)


def apply_algorithm_if_bodies(
    algorithm: Algorithm,
    parent_model: ov.Model,
    parent_graph: NNCFGraph,
    parent_dataset: Dataset,
    subset_size: int,
    current_model_num: int,
    all_models_num: int,
    parent_statistic_points: Optional[StatisticPointsContainer] = None,
) -> Tuple[ov.Model, int]:
    """
    Applies an algorithm recursievley to each bodies of If node.

    :param parent_model: Model to apply algorithm.
    :param parent_graph: Graph of a model.
    :param parent_dataset: Dataset for algorithm.
    :param subset_size: Size of a dataset to use for calibration.
    :param current_model_num: Current model number.
    :param all_models_num: All model numbers.
    :param parent_statistic_points: Statistics points for algorithm.
    :return: A model for every bodies of If nodes the algorithm was applied and the latest model number.
    """
    nncf_logger.info(f"Iteration [{current_model_num}/{all_models_num}] ...")
    quantized_model = algorithm.apply(parent_model, parent_graph, parent_statistic_points, parent_dataset)
    if get_number_if_op(parent_model) == 0:
        return quantized_model, current_model_num
    model_transformer_fp32 = factory.ModelTransformerFactory.create(parent_model)
    for if_node in parent_graph.get_nodes_by_metatypes(OVBackend.if_node_metatypes()):
        parent_model_with_additional_outputs = _add_outputs_before_if_node(
            model_transformer_fp32, parent_model, if_node
        )
        then_model_input_names = OVBackend.get_if_body_input_names(parent_model, if_node, True)
        else_model_input_names = OVBackend.get_if_body_input_names(parent_model, if_node, False)
        if_cond_input_name = OVBackend.get_if_cond_input_name(parent_model_with_additional_outputs, if_node)
        then_dataset, else_dataset = _make_dataset_for_if_bodies(
            factory.EngineFactory.create(parent_model_with_additional_outputs),
            parent_dataset,
            if_cond_input_name,
            then_model_input_names,
            else_model_input_names,
            subset_size,
        )

        then_model = _extract_if_body(model_transformer_fp32, if_node, True)
        then_model = remove_friendly_name_duplicates(then_model)
        else_model = _extract_if_body(model_transformer_fp32, if_node, False)
        else_model = remove_friendly_name_duplicates(else_model)

        then_quantized_model, current_model_num = apply_algorithm_if_bodies(
            algorithm,
            then_model,
            NNCFGraphFactory.create(then_model),
            then_dataset,
            subset_size,
            current_model_num + 1,
            all_models_num,
        )
        else_quantized_model, current_model_num = apply_algorithm_if_bodies(
            algorithm,
            else_model,
            NNCFGraphFactory.create(else_model),
            else_dataset,
            subset_size,
            current_model_num + 1,
            all_models_num,
        )
        model_transformer_int8 = factory.ModelTransformerFactory.create(quantized_model)
        quantized_model = _update_if_body(model_transformer_int8, if_node, True, then_quantized_model)
        model_transformer_int8 = factory.ModelTransformerFactory.create(quantized_model)
        quantized_model = _update_if_body(model_transformer_int8, if_node, False, else_quantized_model)
    return quantized_model, current_model_num


class OVBackend:
    @staticmethod
    def _get_if_body_port_id(if_body_condition: bool):
        """
        Returns port id of a If body based on if_body_condition.

        :param if_body_condition: Condition of If node.
        :return: Port id of body of If node.
        """
        return int(not if_body_condition)

    @staticmethod
    def if_node_metatypes() -> List[OperatorMetatype]:
        """
        Returns metatypes that map to If node.

        :return: Metatypes mapped to If node.
        """
        return [OVIfMetatype]

    @staticmethod
    def get_if_body_input_names(model: ov.Model, if_node: NNCFNode, if_body_condition: bool) -> List[str]:
        """
        Returns input names of If node body based on if_body_condition.
        The order of inputs are in a way that they are passed to the model during inference.

        :param model: Original model.
        :param if_node: If node.
        :param if_body_condition: True for then body, else for else body.
        :return: Input names of If body.
        """
        input_names = []
        name_to_node_mapping = {op.get_friendly_name(): op for op in model.get_ops()}
        ov_node = name_to_node_mapping[if_node.node_name]
        input_indices = [
            desc.input_index
            for desc in ov_node.get_input_descriptions(OVBackend._get_if_body_port_id(if_body_condition))
        ]
        for index in input_indices:
            input_names.append(ov_node.inputs()[index].get_tensor().get_any_name())
        return input_names

    @staticmethod
    def get_if_cond_input_name(model: ov.Model, if_node: NNCFNode) -> str:
        """
        Returns name of condition input of If node.

        :param model: Model.
        :param if_node: If node.
        :return: Name of condition input of If node.
        """
        name_to_node_mapping = {op.get_friendly_name(): op for op in model.get_ops()}
        ov_node = name_to_node_mapping[if_node.node_name]
        return ov_node.inputs()[0].get_tensor().get_any_name()

    @staticmethod
    def create_update_body_command(if_node: NNCFNode, if_body_condition: bool, body: ov.Model) -> OVUpdateIfBodyCommand:
        """
        Returns a command for setting a body of If node by a new one.

        :param if_node: If node.
        :param if_body_condition: Condition of If node.
        :param body: A new body to set.
        :return: Command to update If node body.
        """
        target_point = OVTargetPoint(
            TargetType.LAYER, if_node.node_name, OVBackend._get_if_body_port_id(if_body_condition)
        )
        return OVUpdateIfBodyCommand(target_point, body)

    @staticmethod
    def create_extract_if_body_command(if_node: NNCFNode, if_body_condition: bool) -> OVExtractIfBodyCommand:
        """
        Returns a command for extraction body of If node.
        If if_body_condition is True, extract then body, otherwise - else body.

        :param if_node: If node.
        :param if_body_condition: Condition of body of If node.
        :return: Extracted body of If node.
        """
        return OVExtractIfBodyCommand(if_node.node_name, if_body_condition)

    @staticmethod
    def create_output_insertion_commands_if_node(model: ov.Model, if_node: NNCFNode) -> List[OVOutputInsertionCommand]:
        """
        Returns output insertion commands on If node inputs.

        :param ov.Model model: Model.
        :param NNCFNode if_node: If node.
        :return: Transformation commands to insert outputs before If node.
        """
        assert if_node.metatype == OVIfMetatype
        commands = []
        name_to_node_mapping = {op.get_friendly_name(): op for op in model.get_ops()}
        ov_node = name_to_node_mapping[if_node.node_name]
        for port_id in range(len(ov_node.inputs())):
            commands.append(
                OVOutputInsertionCommand(OVTargetPoint(TargetType.PRE_LAYER_OPERATION, if_node.node_name, port_id))
            )
        return commands



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Any, Dict, List, Tuple, Union

import tensorflow as tf

from nncf.experimental.tensorflow.context import get_current_context
from nncf.experimental.tensorflow.graph.transformations.commands import TFTargetPoint
from nncf.experimental.tensorflow.patch_tf import Hook
from nncf.tensorflow.layers.operation import NNCFOperation

InputSignature = Union[tf.TensorSpec, Dict[str, tf.TensorSpec], Tuple[tf.TensorSpec, ...], List[tf.TensorSpec]]


def _add_names_to_input_signature(input_signature: InputSignature):
    xs = tf.nest.flatten(input_signature)
    ys = []
    for i, spec in enumerate(xs):
        ys.append(tf.TensorSpec.from_spec(spec, name=spec.name if spec.name else f"input_{i}"))

    return tf.nest.pack_sequence_as(input_signature, ys)


class NNCFNetwork(tf.keras.Model):
    """
    Wraps the Keras model.
    """

    def __init__(self, model: tf.keras.Model, input_signature: InputSignature, **kwargs):
        """
        Initializes the NNCF network.

        :param model: Keras model.
        :param input_signature: Input signature of the moodel.
        """
        super().__init__(**kwargs)
        self._model = model
        self._input_signature = _add_names_to_input_signature(input_signature)

        # The `__setattr__` was was overridden inside superclasses.
        # This workaround allows not add dependencies from hooks to the model.
        # See `tensorflow.python.training.tracking.autotrackable.AutoTrackable`
        # class for more details.
        self.__dict__["_pre_hooks"]: Dict[str, List[Hook]] = {}
        self.__dict__["_post_hooks"]: Dict[str, List[Hook]] = {}

    @property
    def nncf_operations(self) -> List[NNCFOperation]:
        """
        Returns list of the NNCF operations which were added to the NNCF network.

        :return: List of the NNCF operations.
        """
        return [op for hook in getattr(self, "_hooks") for op in hook.operations]

    @property
    def input_signature(self) -> InputSignature:
        """
        Returns input signature of the model.

        :return: Input signature of the model.
        """
        return self._input_signature

    def get_nncf_operations_with_params(self) -> List[Tuple[NNCFOperation, Any]]:
        return [(op, hook.get_operation_weights(op.name)) for hook in getattr(self, "_hooks") for op in hook.operations]

    def get_config(self):
        raise NotImplementedError

    def call(self, inputs, **kwargs):
        """
        Calls the model on new inputs and returns the outputs as tensors.
        We call the model inside the tracing context to add the NNCF
        operations to the graph of the model.

        :param inputs: Input tensor, or dict/list/tuple of input tensors.
        :return: A tensor if there is a single output, or a list of tensors
            if there are more than one outputs.
        """
        xs = self._apply_post_hooks_for_inputs(inputs)
        with get_current_context().enter(in_call=True, wrap_ops=True, model=self):
            outputs = self._model(xs, **kwargs)
        return outputs

    def insert_at_point(self, point: TFTargetPoint, ops: List[NNCFOperation]) -> None:
        """
        Inserts the list of the NNCF operations according to the target point.

        :param point: The location where operations should be inserted.
        :param ops: List of the NNCF operarions.
        """
        ops_weights = {op.name: op.create_variables(self) for op in ops}
        hook = Hook(ops, point, ops_weights)
        hooks = getattr(self, "_pre_hooks") if hook.is_pre_hook else getattr(self, "_post_hooks")
        # TODO(andrey-churkin): What we should do if the hook with the same `target_point`
        # already exists inside `hooks`? Is it a valid case?
        hooks.setdefault(hook.target_point.op_name, []).append(hook)

    @property
    def _hooks(self):
        pre_hooks = getattr(self, "_pre_hooks")
        post_hooks = getattr(self, "_post_hooks")
        return itertools.chain(*pre_hooks.values(), *post_hooks.values())

    def _apply_post_hooks_for_inputs(self, inputs):
        """
        Applies post-hooks to inputs.

        :param inputs: Input tensor, or dict/list/tuple of input tensors.
        :return: Modified input tensor, or dict/list/tuple of input tensors.
        """
        input_name_to_post_hook_map = {
            hook.target_point.op_name: hook
            for hook in getattr(self, "_hooks")
            if hook.target_point.op_type_name == "Placeholder"
        }

        if not input_name_to_post_hook_map:
            return inputs

        xs = tf.nest.flatten(inputs)
        ys = tf.nest.flatten(self.input_signature)
        flat_sequence = []
        for input_tensor, input_spec in zip(xs, ys):
            post_hook = input_name_to_post_hook_map[input_spec.name]
            (ans,), _ = post_hook(input_tensor)
            flat_sequence.append(ans)
        return tf.nest.pack_sequence_as(inputs, flat_sequence)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Set, Tuple

import numpy as np

import nncf
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.hardware.config import HWConfig
from nncf.common.quantization.structs import QuantizerConfig
from nncf.experimental.common.tensor_statistics.collectors import AGGREGATORS_MAP
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.openvino.graph.layer_attributes import OVLayerAttributes
from nncf.openvino.graph.metatypes import openvino_metatypes as om
from nncf.openvino.graph.metatypes.groups import OPERATIONS_WITH_WEIGHTS
from nncf.openvino.graph.model_utils import get_start_nodes_for_activation_path_tracing
from nncf.openvino.graph.node_utils import get_weight_channel_axes
from nncf.openvino.graph.transformations.commands import OVConvertInsertionCommand
from nncf.openvino.graph.transformations.commands import OVQuantizerInsertionCommand
from nncf.openvino.graph.transformations.commands import OVTargetPoint
from nncf.openvino.hardware.config import OVHWConfig
from nncf.openvino.quantization.default_quantization import DEFAULT_OV_QUANT_TRAIT_TO_OP_DICT
from nncf.openvino.statistics.collectors import OV_REDUCERS_MAP
from nncf.openvino.statistics.collectors import OVNNCFCollectorTensorProcessor
from nncf.openvino.statistics.statistics import OVMinMaxTensorStatistic
from nncf.parameters import ModelType
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import RangeEstimatorParameters
from nncf.quantization.advanced_parameters import StatisticsType
from nncf.quantization.algorithms.min_max.backend import MinMaxAlgoBackend
from nncf.quantization.fake_quantize import FakeConvertParameters
from nncf.quantization.fake_quantize import FakeQuantizeParameters
from nncf.quantization.range_estimator import AggregatorType


class OVMinMaxAlgoBackend(MinMaxAlgoBackend):
    @property
    def mat_mul_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVMatMulMetatype]

    @property
    def post_processing_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVTopKMetatype, om.OVNonMaxSuppressionMetatype]

    @property
    def conv_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVConvolutionMetatype]

    @property
    def overflow_fix_metatypes(self) -> List[OperatorMetatype]:
        return [
            om.OVConvolutionMetatype,
            om.OVGroupConvolutionMetatype,
            om.OVConvolutionBackpropDataMetatype,
            om.OVGroupConvolutionBackpropDataMetatype,
            om.OVMatMulMetatype,
        ]

    @property
    def add_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVAddMetatype]

    @property
    def group_conv_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVGroupConvolutionMetatype]

    @property
    def shapeof_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVShapeOfMetatype]

    @property
    def dropout_metatypes(self) -> List[OperatorMetatype]:
        return []

    @property
    def read_variable_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVReadValueMetatype]

    @property
    def scaled_dot_product_attention_metatypes(self) -> List[OperatorMetatype]:
        return [om.OVScaledDotProductAttentionMetatype]

    @property
    def scales_unification_map(self) -> Dict[OperatorMetatype, OperatorMetatype]:
        return {om.OVConcatMetatype: self.overflow_fix_metatypes}

    @property
    def hw_config(self) -> HWConfig:
        return OVHWConfig

    @property
    def quant_trait_op_dict(self) -> Dict[int, OperatorMetatype]:
        return DEFAULT_OV_QUANT_TRAIT_TO_OP_DICT

    @staticmethod
    def get_start_nodes_for_activation_path_tracing(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return get_start_nodes_for_activation_path_tracing(nncf_graph)

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> OVTargetPoint:
        return OVTargetPoint(target_type, target_node_name, port_id)

    @staticmethod
    def create_quantizer_insertion_command(
        nncf_graph: NNCFGraph,
        target_point: OVTargetPoint,
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> OVQuantizerInsertionCommand:
        return OVQuantizerInsertionCommand(target_point, parameters)

    @staticmethod
    def create_unified_scales_quantizers_insertion_commands(
        nncf_graph: NNCFGraph,
        target_points: List[OVTargetPoint],
        quantizer_config: QuantizerConfig,
        parameters: FakeQuantizeParameters,
    ) -> List[OVQuantizerInsertionCommand]:
        return [OVQuantizerInsertionCommand(target_point, parameters) for target_point in target_points]

    @staticmethod
    def create_convert_insertion_command(
        target_point: OVTargetPoint,
        parameters: FakeConvertParameters,
    ) -> OVQuantizerInsertionCommand:
        return OVConvertInsertionCommand(target_point, parameters)

    @staticmethod
    def unify_statistics(statistics: List[OVMinMaxTensorStatistic]) -> OVMinMaxTensorStatistic:
        max_values, min_values = [], []
        for statistic in statistics:
            max_values.append(np.array(statistic.max_values).flatten())
            min_values.append(np.array(statistic.min_values).flatten())
        max_values = np.max(max_values, axis=0)
        min_values = np.min(min_values, axis=0)
        return OVMinMaxTensorStatistic(min_values=min_values, max_values=max_values)

    @staticmethod
    def get_target_point_shape(nncf_graph: NNCFGraph, node: NNCFNode, target_point: OVTargetPoint) -> Tuple[int, ...]:
        if target_point.is_weight_target_point():
            return node.layer_attributes.constant_attributes[target_point.port_id]["shape"]
        if target_point.type == TargetType.PRE_LAYER_OPERATION:
            return nncf_graph.get_input_edges(node)[target_point.port_id].tensor_shape
        elif target_point.type == TargetType.POST_LAYER_OPERATION:
            return nncf_graph.get_output_edges(node)[target_point.port_id].tensor_shape
        raise NotImplementedError(f"Unsupported target point type {target_point.type}.")

    @staticmethod
    def get_weight_quantization_axes(node: NNCFNode, target_point: OVTargetPoint) -> Tuple[int]:
        return tuple(get_weight_channel_axes(node))

    @staticmethod
    def get_statistic_collector(
        range_estimator_params: RangeEstimatorParameters,
        use_abs_max: bool,
        reduction_axes: Optional[Tuple[int, ...]],
        aggregation_axes: Optional[Tuple[int, ...]],
        inplace: bool,
        num_samples: Optional[int] = None,
    ) -> TensorCollector:
        collector = TensorCollector(OVMinMaxTensorStatistic)
        for params, container_key in zip(
            [range_estimator_params.min, range_estimator_params.max],
            [OVMinMaxTensorStatistic.MIN_STAT, OVMinMaxTensorStatistic.MAX_STAT],
        ):
            if params.statistics_type not in OV_REDUCERS_MAP:
                raise nncf.InternalError(
                    f"Statistic type: {params.statistics_type} is not supported for OpenVino PTQ backend yet."
                )
            if params.aggregator_type not in AGGREGATORS_MAP:
                raise nncf.InternalError(
                    f"Aggregator type: {params.aggregator_type} is not supported for OpenVino PTQ backend yet."
                )
            kwargs = {"reduction_axes": reduction_axes, "inplace": inplace}
            if params.statistics_type in [StatisticsType.QUANTILE, StatisticsType.ABS_QUANTILE]:
                if container_key == OVMinMaxTensorStatistic.MIN_STAT:
                    quantile = params.quantile_outlier_prob
                else:
                    quantile = 1 - params.quantile_outlier_prob
                kwargs.update({"quantile": [quantile]})
            # TODO(dlyakhov): merge two quantile aggregators in one
            statistic_type = params.statistics_type
            if use_abs_max and statistic_type == StatisticsType.MAX:
                statistic_type = StatisticsType.ABS_MAX
            reducer = OV_REDUCERS_MAP[statistic_type](**kwargs)

            kwargs = {
                "num_samples": num_samples,
                "aggregation_axes": aggregation_axes,
                "tensor_processor": OVNNCFCollectorTensorProcessor,
            }
            if params.aggregator_type in [AggregatorType.MEAN_NO_OUTLIERS, AggregatorType.MEDIAN_NO_OUTLIERS]:
                kwargs.update({"quantile": params.quantile_outlier_prob})
            aggregator = AGGREGATORS_MAP[params.aggregator_type](**kwargs)

            collector.register_statistic_branch(container_key, reducer, aggregator)
        return collector

    @staticmethod
    def get_weight_tensor_port_ids(node: NNCFNode) -> List[Optional[int]]:
        return node.layer_attributes.get_const_port_ids()

    @staticmethod
    def get_ignored_metatypes(model_type: ModelType, device: TargetDevice) -> List[OperatorMetatype]:
        types = []
        if model_type == ModelType.TRANSFORMER:
            types = [
                om.OVAddMetatype,
                om.OVPowerMetatype,
                om.OVSqueezeMetatype,
                om.OVSubtractMetatype,
                om.OVAvgPoolMetatype,
                om.OVReduceMeanMetatype,
                om.OVReduceL2Metatype,
                om.OVSumMetatype,
                om.OVSquaredDifferenceMetatype,
                om.OVMVNMetatype,
                om.OVBatchNormMetatype,
                om.OVDivideMetatype,
                om.OVSqrtMetatype,
                om.OVMaximumMetatype,
            ]
            if device != TargetDevice.CPU_SPR:
                types.append(om.OVMultiplyMetatype)
        return types

    @staticmethod
    def get_ignored_names_by_layer_attributes(nncf_graph: NNCFGraph) -> List[str]:
        ignored_names = []
        target_nodes = nncf_graph.get_nodes_by_metatypes([om.OVGRUSequenceMetatype])
        for node in target_nodes:
            if (
                isinstance(node.layer_attributes, OVLayerAttributes)
                and node.layer_attributes.input_attributes["linear_before_reset"]
            ):
                ignored_names.append(node.node_name)
        return ignored_names

    @staticmethod
    def get_weight_nodes(nncf_graph: NNCFGraph) -> List[NNCFNode]:
        return [
            node
            for node in nncf_graph.get_all_nodes()
            if isinstance(node.layer_attributes, OVLayerAttributes) and node.metatype in OPERATIONS_WITH_WEIGHTS
        ]

    @staticmethod
    def get_weight_name(nncf_graph: NNCFGraph, target_point: OVTargetPoint) -> str:
        node = nncf_graph.get_node_by_name(target_point.target_node_name)
        return node.layer_attributes.constant_attributes[target_point.port_id]["name"]

    @staticmethod
    def should_quantize_weight(weight_name: str, quantized_weight_names: Set[str]) -> bool:
        return True
